header {* Isabelle como un lenguaje funcional *}

(*<*)
theory Cap_1 
imports Main LaTeXsugar OptionalSugar
begin
(*>*)

section {* Introducción *}

text {*
  \begin{nota}
    Esta notas son una introducción a la demostración asistida utilizando
    el sistema Isabelle/HOL/Isar. La versión de Isabelle utilizada es la
    2009-2. 
  \end{nota}

  \begin{nota}
    Un \textbf{lema} introduce una proposición seguida de una
    demostración. Isabelle dispone de varios procedimientos automáticos para
    generar demostraciones, uno de los cuales es el de simplificación (llamado
    \textit{simp}). El procedimiento \textit{simp} aplica un conjunto de reglas
    de reescritura que inicialmente contiene un gran número de reglas relativas
    a los objetos definidos. El ejemplo del lema más trivial es el siguiente
  \end{nota}
*}

lemma elMasTrivial: "True" 
by simp

text {* 
  En este capítulos se presenta el lenguaje funcional que está incluido en
  Isabelle. El lenguaje funcional es muy parecido al ML estándard.
*}

section {* Números naturales, enteros y booleanos *}

text {*
  \begin{nota}[Números naturales]
    \mbox{} 
    \begin{itemize}
    \item En Isabelle están definidos los número naturales con la sintaxis de
      Peano usando dos constructores: @{text 0} (cero) y @{text "Suc n"} (el
      sucesor de @{text n}). 
    \item Los números como el @{text 1}  son abreviaturas de los
      correspondientes en la notación de Peano, en este caso @{text "Suc 0"}.
    \item El tipo de los números naturales es @{text nat}. 
    \end{itemize}
  \end{nota}

  \begin{lema}[Ejemplo de simplificación de números naturales]
  El siguiente del $0$ es el $1$.
  \end{lema}
*}

lemma "Suc 0 = 1" 
by simp

text {* 
  \begin{nota}[Suma y producto de números naturales]
  En Isabelle están definida la suma y el producto de números naturales:
  \begin{itemize}
    \item @{text "x+y"} es la suma de @{text x} e @{text y} 
    \item @{text "x*y"} es el producto de @{text x} e @{text y} 
  \end{itemize}
  \end{nota}

  \begin{lema}[Ejemplo de suma]
  La suma de los números naturales 1 y 2 es el número natural 3.
  \end{lema}
*}

lemma "1 + 2 = (3::nat)" 
by simp

text {* 
  \begin{nota}[Epecificación de tipo]
    La notación del par de dos puntos se usa para asignar un tipo a un término
    (por ejemplo, \texttt{3::nat} significa que se considera que \texttt{3} es
    un número natural).   
  \end{nota}

  \begin{lema}[Ejemplo de producto]
  El producto de los números naturales 2 y 3 es el número natural 6.
  \end{lema}
*}

lemma "2 * 3 = (6::nat)" 
by simp

text {* 
  \begin{nota}[División de números naturales]
  En Isabelle está definida la división de números naturales: @{term "n div m"}
  es el mayor número natural que multiplicado por @{term "m"} es menor o igual
  que @{term "n"}.
  \end{nota}

  \begin{lema}[Ejemplo de división]
  La división natural de 7 entre 3 es 2.
  \end{lema}
*}

lemma "7 div 3 = (2::nat)" 
by simp

text {* 
  \begin{nota}[Resto de división de números naturales]
  En Isabelle está definida el resto de división de números naturales: 
  @{term "n mod m"} es el resto de dividir @{term "n"} entre @{term "m"}.
  \end{nota}

  \begin{lema}[Ejemplo de resto]
  El resto de dividir 7 entre 3 es 1.
  \end{lema}
*}

lemma "7 mod 3 = (1::nat)" 
by simp

text {* 
  \begin{nota}[Números enteros]
  En Isabelle también están definidos los números enteros. El tipo de los
  enteros se representa por \textit{int}.
  \end{nota}

  \begin{lema}[Ejemplo de operación con enteros]
  La suma de 1 y -2 es el número entero -1.
  \end{lema}
*}

lemma "1 + -2 = (-1::int)" 
by simp

text {* 
  \begin{nota}[Sobrecarga]
  Los numerales están sobrecargados. Por ejemplo, el '1' puede ser un natural o
  un entero, dependiendo del contexto. Isabelle resuelve ambigüedades mediante
  inferencia de tipos. A veces, es necesario usar declaraciones de tipo para
  resolver la ambigüedad.  
  \end{nota}

  \begin{nota}[Booleanos, conectivas y cuantificadores]
  En Isabelle están definidos los valores booleanos 
  @{text "True, False"}, las conectivas @{text "\<not>, \<and>, \<or>, \<longrightarrow>, \<leftrightarrow>"} y los
  cuantificadores @{text "\<forall>, \<exists>"}. El tipo de los booleanos es \textit{bool}. 
  \end{nota}

  \begin{lema}[Ejemplos de evaluaciones booleanas]
    \mbox{}
    \begin{enumerate}
    \item La conjunción de dos fórmulas verdaderas es verdadera.
    \item La conjunción de un fórmula verdadera y una falsa es falsa.
    \item La disyunción de una fórmula verdadera y una falsa es verdadera.
    \item La disyunción de dos fórmulas falsas es falsa.
    \item La negación de una fórmula verdadera es falsa.
    \item Una fórmula falsa implica una fórmula verdadera.
    \item Todo elemento es igual a sí mismo.
    \item Existe un elemento igual a 1.
    \end{enumerate}
  \end{lema}
*}

lemma "True \<and> True = True" 
by simp

lemma "True \<and> False = False" 
by simp
 
lemma "True \<or> False = True" 
by simp

lemma "False \<or> False = False" 
by simp

lemma "\<not> True = (False::bool)" 
by simp

lemma "False \<longrightarrow> True" 
by simp

lemma "\<forall> x. x = x" 
by simp

lemma "\<exists> x. x = 1" 
by simp

section {* Definiciones no recursivas *}

text {*
  \begin{definicion}[Ejemplo de definición no recursiva]
  La disyunción exclusiva de $A$ y $B$ se verifica si una es verdadera y la
  otra no lo es.
  \end{definicion}
*}

definition xor :: "bool \<Rightarrow> bool \<Rightarrow> bool" where
  "xor A B \<equiv> (A \<and> \<not> B) \<or> (\<not> A \<and> B)"

text {* 
  \begin{lema}[Ejemplo de demostración con definiciones no recursivas]
  La disyunción exclusiva de dos fórmulas verdaderas es falsa.
  \end{lema}

  \begin{demostracion}
  Por simplificación, usando la definición de la disyunción exclusiva.
  \end{demostracion}
*}

lemma "xor True True = False"
by (simp  add: xor_def)

text {* 
  \begin{nota}[Ejemplo de ampliación de las reglas de simplificación]
  Se añade la definición de la disyunción exlusiva al conjunto de reglas de
  simplificación automáticas.
  \end{nota}
*}

declare xor_def[simp]

section {* Definiciones locales *}

text {*
  \begin{nota}[Variables locales]
  Se puede asignar valores a variables locales mediante 'let' y usarlo en las 
  expresiones dentro de 'in'. 
  \end{nota}

  \begin{lema}[Ejemplo de entorno local]
  Sea $x$ el número natural 3. Entonces @{text "x \<times> x = 9"}.
  \end{lema}
*}

lemma "(let x = 3::nat in x * x = 9)" 
by simp

section {* Pares *}

text {*
  \begin{nota}[Pares]
  \mbox{}
  \begin{itemize}
  \item Un par se representa escribiendo los elementos entre paréntesis y
    separados por coma.
  \item El tipo de los pares es el producto de los tipos.
  \item La función \textit{fst} devuelve el primer elemento de un par y la
    \textit{snd} el segundo.
  \end{itemize}
  \end{nota}

  \begin{lema}[Ejemplo de uso de pares]
  Sea $p$ el par de números naturales $(2,3)$. La suma del primer elemento de
  $p$ y $1$ es igual al segundo elemento de $p$.
  \end{lema}
*}

lemma "let p = (2,3)::nat \<times> nat in fst p + 1 = snd p" 
by simp

section {* Listas *}

text {*
  \begin{nota}[Construcción de listas]
    \mbox{}
    \begin{itemize}
    \item Una lista se representa escribiendo los elementos entre corchetes y
      separados por coma.
    \item La lista vacía se representa por \verb|[]|.
    \item Todos los elementos de una lista tienen que ser del mismo tipo.
    \item El tipo de las listas de elementos del tipo \verb|a| es \verb|a list|.
    \item El término \verb|a#l| representa la lista obtenida añadiendo el
      elemento \verb|a| al principio de la lista \verb|l|.
    \end{itemize}
  \end{nota}

  \begin{lema}[Ejemplo de construcción de listas]
  La lista obtenida añadiendo sucesivamente a la lista vacía los elementos 3,
  2 y 1 es [1,2,3]. 
  \end{lema}
*}

lemma "1#(2#(3#[])) = [1,2,3]" 
by simp

text {* 
  \begin{nota}[Primero y resto]
    \mbox{}
    \begin{itemize}
    \item \verb|hd l| es el primer elemento de la lista \verb|l|.
    \item \verb|tl l| es el resto de la lista \verb|l|.
    \end{itemize}
  \end{nota}

  \begin{lema}[Ejemplo de cálculo con listas]
  Sea \verb|l| la lista de números naturales \verb|[1,2,3]|. Entonces, el
  primero de \verb|l| es 1 y el resto de \verb|l| es \verb|[2,3]|.  
  \end{lema}
*}

lemma "let l = [1,2,3]::(nat list) in hd l = 1 \<and> tl l = [2,3]" 
by simp

text {* 
  \begin{nota}[Longitud]
  \verb|length l| es la longitud de la lista \verb|l|.
  \end{nota}

  \begin{lema}[Ejemplo de cálculo de longitud]
  La longitud de la lista [1,2,3] es 3.
  \end{lema}
*}

lemma "length [1,2,3] = 3" 
by simp

text {* 
  \begin{nota}[Referencias sobre listas]
  En la sesión 38 de
  \href{http://isabelle.informatik.tu-muenchen.de/library/HOL/outline.pdf}
       {``HOL: The basis of Higher-Order Logic''} 
  se encuentran  más definiciones y propiedades de las listas.
  \end{nota}
*}

section {* Registros *}

text {*
  \begin{nota}[Registro]
  Un registro es una colección de campos y valores. 
  \end{nota}

  \begin{definicion}[Ejemplo de definición de registro]
  Los puntos del plano pueden representarse mediante registros con dos campos,
  las coordenadas, con valores enteros.
  \end{definicion}
*}

record punto = 
  coordenada_x :: int
  coordenada_y :: int

text {* 
  \begin{definicion}[Ejemplo de definición de un registro]
  El punto \verb|pt| tiene de coordenadas 3 y 7.  
  \end{definicion}
*}

definition pt :: punto where
  "pt \<equiv> \<lparr>coordenada_x = 3, coordenada_y = 7\<rparr>"

text {* 
  \begin{lema}[Ejemplo de propiedad de registro]
  La coordenada x del punto \verb|pt| es 3.  
  \end{lema}
*}

lemma "coordenada_x pt = 3" 
by (simp add: pt_def)

text {* 
  \begin{lema}[Ejemplo de actualización de un registro]
  Sea \verb|pt2| el punto obtenido a partir del punto \verb|pt| cambiando el
  valor de su coordenada x por 4. Entonces la coordenada x del punto \verb|pt2|
  es 4. 
  \end{lema}
*}

lemma "let pt2=pt\<lparr>coordenada_x:=4\<rparr> in coordenada_x (pt2) = 4" 
by (simp add: pt_def)

section {* Funciones anónimas *}

text {*
  \begin{nota}[Funciones anónimas]
  En Isabelle pueden definirse funciones anónimas.  
  \end{nota}

  \begin{lema}[Ejemplo de uso de funciones anónimas]
  El valor de la función que a un número le asigna su doble aplicada a 1 es 2.  
  \end{lema}
*}

lemma "(\<lambda> x. x + x) 1 = (2::nat)" 
by simp

section {* Condicionales *}

text {*
  \begin{definicion}[Ejemplo con el condicional \textit{if}]
  El valor absoluto del entero $x$ es $x$, si @{text "x \<ge> 0"} y es $-x$ en caso
  contrario.    
  \end{definicion}
*}

definition absoluto :: "int \<Rightarrow> int" where
  "absoluto x \<equiv> (if x \<ge> 0 then x else -x)"

text {* 
  \begin{lema}[Ejemplo de simplificación con el condicional \textit{if}]
  El valor absoluto de -3 es 3.  
  \end{lema}
*}

lemma "absoluto(-3) = 3"
by (simp add:absoluto_def) 

text {* 
  \begin{definicion}[Ejemplo con el condicional \textit{case}]
  Un número natural $n$ es un sucesor si es de la forma @{term "Suc m"}.
  \end{definicion}
*}

definition es_sucesor :: "nat \<Rightarrow> bool" where
  "es_sucesor n \<equiv>
  (case n of 
    0     \<Rightarrow> False 
  | Suc m \<Rightarrow> True)"

text {* 
  \begin{lema}[Ejemplo de simplificación con el condicional \textit{case}]
  El número 3 es sucesor.  
  \end{lema}
*}

lemma "es_sucesor 3"
by (simp add: es_sucesor_def)

section {* Tipos de datos y recursión primitiva *}

text {*
  \begin{definicion}[Ejemplo de definición de tipo de dato recursivo]
  Una lista de elementos de tipo \verb|a| es la lista \verb|Vacia| o se
  obtiene añadiendo, con \verb|ConsLista|, un elemento de tipo \verb|a| a una
  lista de elementos de tipo \verb|a|.  
  \end{definicion}
*}

datatype 'a Lista = Vacia | ConsLista 'a "'a Lista"

text {* 
  \begin{definicion}[Ejemplo de definición primitiva recursiva]
  \verb|conc xs ys| es la concatenación de las lista \verb|xs| e \verb|ys|.  
  \end{definicion}
*}

primrec conc :: "'a Lista \<Rightarrow> 'a Lista \<Rightarrow> 'a Lista" where
  "conc Vacia ys = ys"
| "conc (ConsLista x xs) ys = ConsLista x (conc xs ys)"

text {* 
  \begin{lema}[Ejemplo de simplificación con tipo de dato recursivo]
  La concatenación de la lista formada por 1 y 2 con la lista formada por el 3
  es la lista cuyos elementos son 1,2 y 3.
  \end{lema}
*}

lemma "conc (ConsLista 1 (ConsLista 2 Vacia)) (ConsLista 3 Vacia) =
       (ConsLista 1 (ConsLista 2 (ConsLista 3 Vacia)))"
by simp

text {* 
  \begin{ejercicio}[Ejemplo de definición primitiva recursiva sobre los
  naturales] 
  Definir una función que sume los primeros n números naturales y usarla para
  comprobar que la suma de los 3 primeros números naturales es 6.
  \end{ejercicio}
*}

primrec suma :: "nat \<Rightarrow> nat" where
  "suma 0 = 0"
| "suma (Suc m) = (Suc m) + suma m"

lemma "suma 3 = 6"
by (simp add: suma_def) 

(*<*)
end
(*>*)

