header {* El lenguaje de demostración Isa *}

(*<*)
theory Cap_2
imports Main LaTeXsugar OptionalSugar 
begin
(*>*)

text {*
  Este capítulo describe los elementos básicos del lenguaje de demostración
  Isar (\emph{Intelligible semi--automated reasoning}).
*}

section {* Panorama de la sintaxis (simplificada) de Isar *}

text {* 
 \begin{nota}[Representación de lemas (y teoremas)]
  \mbox{}
  \begin{itemize}
  \item Un \textbf{lema} (o \textbf{teorema}) comienza con una
    \textbf{etiqueta} seguida por algunas \textbf{premisas} y una
    \textbf{conclusión}.
  \item Las premisas se introducen con la palabra \textbf{assumes} y se separan
    con \textbf{and}.
  \item Cada premisa puede etiquetarse para referenciarse en la demostración.
  \item La conclusión se introduce con la palabra \textbf{shows}.
  \end{itemize}
  \end{nota}

  \begin{nota}[Gramática (simplificada) de las demostraciones en Isar]
  \mbox{} \newline
  \begin{tabular}{lll}
  demostración 
    & ::=      & \textbf{proof} método declaración* \textbf{qed} \\
    & \textbar & \textbf{by} método \\
  declaración
    & ::=      & \textbf{fix} variable+ \\
    & \textbar & \textbf{assume} proposición+ \\
    & \textbar & (\textbf{from} hecho+)? 
                 \textbf{have} proposición+ demostración \\
    & \textbar & (\textbf{from} hecho+)? 
                 \textbf{show} proposición+ demostración \\
  proposición  
    & ::=      & (etiqueta\textbf{:})? cadena \\
  hecho        
    & ::=      & etiqueta \\
  método
    & ::=      & \verb|-| \\
    & \textbar & \textbf{this} \\
    & \textbar & \textbf{rule} hecho \\ 
    & \textbar & \textbf{simp}  \\
    & \textbar & \textbf{blast} \\
    & \textbar & \textbf{auto} \\
    & \textbar & \textbf{induct} variable \\ 
    & \textbar & ... 
  \end{tabular}

  La declaración \textbf{show} demuestra la conclusión de la demostración
  mientras que la declaración \textbf{have} demuestra un resultado
  intermedio.     
  \end{nota}
*}

section "Razonamiento proposicional"

text {* 
  \begin{nota}[Regla de introducción de la conjunción]
    \begin{equation}\notag
      (\textit{conjI})\;@{thm [mode=Rule] conjI }
    \end{equation}
  \end{nota}

  \begin{lema}[Ejemplo de introducción de conjunción con razonamiento
    progresivo] 
    \mbox{} \newline \hspace*{1cm}
    @{text "P, Q \<turnstile> P \<and> (Q \<and> P)"}.  
  \end{lema}

  \begin{demostracion}
  Estamos suponiendo
  \begin{equation}\label{P}
    P
  \end{equation}
  y
  \begin{equation}\label{Q}
    Q
  \end{equation}
  De \ref{Q} y \ref{P}, por introducción de la conjunción, se tiene
  \begin{equation}\label{QyP}
    Q \land P
  \end{equation}
   De \ref{P} y \ref{QyP}, por introducción de la conjunción, se tiene
   @{text "P \<and> (Q \<and> P)"}.
  \end{demostracion}
*}

lemma conj2: 
  assumes p: P and q: "Q" 
  shows "P \<and> (Q \<and> P)"
proof -
  from q p have qp: "Q \<and> P" by (rule conjI)
  from p qp show "P \<and> (Q \<and> P)" by (rule conjI)
qed

text {* 
  \begin{nota}[Razonamiento progresivo y regresivo]
  \mbox{}
  \begin{itemize}
  \item Isabelle soporta \emph{razonamiento progresivo}. La anterior
    demostración es una muestra.
  \item Isabelle soporta \emph{razonamiento regresivo}. La siguiente
    demostración es una muestra.
  \end{itemize}
  \end{nota}

  \begin{lema}[Ejemplo de introducción de la conjunción con razonamiento
    regresivo] 
    \mbox{} \newline \hspace*{1cm}
    @{text "P, Q \<turnstile> P \<and> (Q \<and> P)"}
  \end{lema}

  \begin{demostracion}
  Estamos suponiendo
  \begin{equation}
    \label{l2a}
    P
  \end{equation}
  y
  \begin{equation}
    \label{l2b}
    Q
  \end{equation}

  Para demostrar el lema, por introducción de la conjunción, basta probar
  \begin{equation}
    \label{l2c}
    P
  \end{equation}
  y
  \begin{equation}
    \label{l2d}
    Q \land P
  \end{equation}

  La condición \ref{l2c} se tiene por la hipótesis \ref{l2a}. Para demostrar
  la condición \ref{l2d}, por introducción de la conjunción, basta probar
  \begin{equation}
    \label{l2e}
    Q
  \end{equation}
  y
  \begin{equation}
    \label{l2f}
    P
  \end{equation}
  La condición \ref{l2e} se tiene por la hipótesis \ref{l2b} y la condición
  \ref{l2f} se tiene por la hipótesis \ref{l2a}. 
  \end{demostracion}
*}

lemma 
  assumes p: "P" and q: "Q" 
  shows "P \<and> (Q \<and> P)"
proof (rule conjI)
  from p show "P" by this
next
  show "Q \<and> P"
  proof (rule conjI)
    from q show "Q" by this
  next
    from p show "P" by this
  qed
qed

text {* 
  \begin{nota}[El método \textit{this}]
    El método \textit{this} demuestra el objetivo usando el hecho actual (es
    decir, el de la cláusula \textbf{from}). 
  \end{nota}

  \begin{nota}[Reglas de eliminación de la conjunción]
    \begin{equation}\notag
     (\textit{conjunct1})\;@{thm [mode=Rule] conjunct1 }\qquad
     (\textit{conjunct2})\;@{thm [mode=Rule] conjunct2 }
    \end{equation}
  \end{nota}

  \begin{nota}[Regla de introducción de la implicación]
    \begin{equation}\notag
      (\textit{impI})\;@{thm [mode=Rule] impI }
    \end{equation}
  \end{nota}

  \begin{lema}[Ejemplo de razonamiento híbrido]
    Sean $a$ y $b$ dos números naturales. Si $0<a$ y $a<b$, entonces 
    @{text "a * a < b * b"}    
  \end{lema}
*}

lemma
  fixes a b :: "nat"
  shows "0 < a \<and> a < b \<longrightarrow> a * a < b * b"
proof (rule impI)
  assume x: "0 < a \<and> a < b"
  from x have za: "0 < a" by (rule conjunct1)
  from x have ab: "a < b" by (rule conjunct2)
  from za ab have aa: "a*a < a*b" by simp
  from ab have bb: "a*b < b*b" by simp
  from aa bb show "a*a < b*b" by arith
qed

text {* 
  \begin{nota}[Modus ponens]
    \begin{equation}\notag
      (\textit{mp})\;@{thm [mode=Rule] mp }
    \end{equation}
  \end{nota}

  \begin{nota}[Reglas de introducción de la disyunción]
    \begin{equation}\notag
     (\textit{disjI1})\;@{thm [mode=Rule] disjI1 }\qquad
     (\textit{disjI2})\;@{thm [mode=Rule] disjI2 }
    \end{equation}
  \end{nota}

  \begin{nota}[Regla de eliminación de la disyunción]
    \begin{equation}\notag
     (\textit{disjE})\;@{thm [mode=Rule] disjE }
    \end{equation}
  \end{nota}

  \begin{lema}[Razonamiento por casos] 
    \mbox{} \newline \hspace*{1cm}
    @{text "A \<or> B, A \<longrightarrow> C, B \<longrightarrow> C \<turnstile> C"}
  \end{lema}
*}

lemma 
  assumes ab: "A \<or> B" and ac: "A \<longrightarrow> C" and bc: "B \<longrightarrow> C"
  shows "C"
proof -
  note ab
  moreover { 
    assume a: "A" 
    from ac a have "C" by (rule mp) } 
  moreover { 
    assume b: "B" 
    from bc b have "C" by (rule mp) }
  ultimately show "C" by (rule disjE)
qed

text {* 
  \begin{nota}[Resumen de reglas proposicionales]
   $$\begin{array}{lll}
      \textit{TrueI}            && @{thm TrueI } \\ 
      \textit{FalseE}           && @{thm FalseE } \\ \\ 
      \textit{conjI}            && @{thm conjI } \\
      \textit{conjunct1}        && @{thm conjunct2 } \\
      \textit{conjE}            && @{thm conjE } \\ \\
      \textit{disjI1}           && @{thm disjI1 } \\
      \textit{disjI2}           && @{thm disjI2 } \\
      \textit{disjE}            && @{thm disjE } \\ \\
      \textit{notI}             && @{thm notI } \\
      \textit{notE}             && @{thm notE } \\ \\
      \textit{impI}             && @{thm impI } \\ 
      \textit{impE}             && @{thm impE } \\ 
      \textit{mp}               && @{thm mp } \\ \\
      \textit{iff}              && @{thm iff } \\
      \textit{iffI}             && @{thm iffI } \\
      \textit{iffD1}            && @{thm iffD1 } \\
      \textit{iffD2}            && @{thm iffD2 } \\
      \textit{iffE}             && @{thm iffE } \\ \\
      \textit{ccontr}           && @{thm ccontr } \\
      \textit{classical}        && @{thm classical } \\
      \textit{exlude\_middle}   && @{thm "excluded_middle"} \\
      \textit{disjCI}           && @{thm disjCI } \\
      \textit{impCE}            && @{thm impCE } \\
      \textit{iffCE}            && @{thm iffCE } \\
      \textit{notnotD}          && @{thm notnotD } \\
      \textit{swap}             && @{thm swap } \\ \\
     \end{array}$$
  \end{nota}

  \begin{nota}[Referencia de reglas de inferencia]
    Más información sobre las reglas de inferencia se encuentra en la sección
    2.2 de 
    \href{http://www.cl.cam.ac.uk/research/hvg/Isabelle/dist/Isabelle/doc/logics-HOL.pdf}{Isabelle's Logics: HOL}.
  \end{nota}
*}

section {* Atajos de Isar *}

text {*
  \begin{nota}[Atajos de Isar]
  Isar tiene muchos atajos, como los siguientes:

  \begin{tabular}{lll}
  \emph{this}   & (éste)      &= el hecho probado en la declaración anterior \\
  \textbf{then} & (entonces)       &= \textbf{from this}\\
  \textbf{hence}& (por lo tanto)   &= \textbf{then have}\\
  \textbf{thus} & (de esta manera) &= \textbf{then show}\\
  \textbf{with} hecho+ & (con)     &= \textbf{from} hecho+ \textbf{and} \emph{this}\\
  \textbf{.}    & (por ésto)       &= \textbf{by} \emph{this}\\
  \textbf{..}   & (trivialmente)&= \textbf{by} regla (donde Isabelle adivina la
  regla)
  \end{tabular}
  \end{nota}

  \begin{nota}[Razonamiento acumulativo]
    Una sucesión de hechos que se van a usar como premisa en una declaración
    puede agruparse usando \textbf{moreover} (además) y usarse en la
    declaración usando \textbf{ultimately} (finalmente).
  \end{nota} 

  \begin{lema}[Ejemplo de uso de atajos y razonamiento acumulativo]
    \mbox{} \newline \hspace*{1cm}
    @{text "A \<and> B \<turnstile> B \<and> A"}.
  \end{lema}
*}

lemma "A \<and> B \<longrightarrow> B \<and> A"
proof (rule impI)
  assume ab: "A \<and> B"
  hence "B" by (rule conjunct2)
  moreover from ab have "A" ..
  ultimately show "B \<and> A" by (rule conjI)
qed

section {* Cuantificadores universal y existencial *}

text {* 
  \begin{nota}[Reglas del cuantificador universal]
    \begin{equation}\notag
     (\textit{allI})\;@{thm [mode=Rule] allI }\qquad
     (\textit{allE})\;@{thm [mode=Rule] allE }
    \end{equation}
  En la regla \textit{allI} la nueva variable se introduce mediante la palabra
  \textbf{fix}.
  \end{nota}

  \begin{lema}[Ejemplo con  cuantificadores universales]
    \mbox{} \newline \hspace*{1cm}
    @{text "\<forall>x. P \<longrightarrow> Q x \<turnstile> P \<longrightarrow> (\<forall>x. Q x)"}
  \end{lema}
*}

lemma
  assumes a: "\<forall> x. P \<longrightarrow> Q x"
  shows "P \<longrightarrow> (\<forall> x. Q x)"
proof (rule impI)
  assume p: "P"
  show "\<forall> x. Q x"
  proof (rule allI)
    fix x
    from a have pq: "P \<longrightarrow> Q x" by (rule allE)
    from pq p show "Q x" by (rule mp)
  qed
qed

text {* 
  \begin{nota}[Reglas del cuantificador existencial]
    \begin{equation}\notag
     (\textit{exI})\;@{thm [mode=Rule] exI }\qquad
     (\textit{exE})\;@{thm [mode=Rule] exE }
    \end{equation}
  En la regla \textit{exE} la nueva variable se introduce mediante la
  declaración \newline
  '\textbf{obtain} \dots \textbf{where} \dots \textbf{by} (\emph{rule exE})'. 
  \end{nota}

  \begin{lema}[Ejemplo con cuantificador existencial y demostración progresiva]
    \mbox{} \newline \hspace*{1cm}
    @{text "\<exists>x. P \<and> Q(x) \<turnstile> P \<and> (\<exists>x. Q(x))"}
  \end{lema}
*}

lemma
  assumes e: "\<exists> x. P \<and> Q(x)"
  shows "P \<and> (\<exists> x. Q(x))"
proof -
  from e obtain x where f: "P \<and> Q(x)" by (rule exE) 
  from f have p: "P" by (rule conjunct1)
  from f have q: "Q(x)" by (rule conjunct2)
  from q have eq: "\<exists> x. Q(x)" by (rule exI)
  from p eq show "P \<and> (\<exists> x. Q(x))" by (rule conjI)
qed

text {* 
  \begin{lema}[Ejemplo con cuantificador existencial y demostración progresiva
    automática] 
    \mbox{} \newline \hspace*{1cm}
    @{text "\<exists>x. P \<and> Q(x) \<turnstile> P \<and> (\<exists>x. Q(x))"}
  \end{lema}
*}

lemma
  assumes e: "\<exists> x. P \<and> Q(x)"
  shows "P \<and> (\<exists> x. Q(x))"
proof -
  from e obtain x where f: "P \<and> Q(x)" ..
  from f have p: "P" ..
  from f have q: "Q(x)" ..
  from q have eq: "\<exists> x. Q(x)" ..
  from p eq show "P \<and> (\<exists> x. Q(x))" ..
qed

text {* 
  \begin{lema}[Ejemplo con cuantificador existencial y demostración regresiva]
    \mbox{} \newline \hspace*{1cm} 
    @{text "\<exists>x. P \<and> Q(x) \<turnstile> P \<and> (\<exists>x. Q(x))"}
  \end{lema}
*}

lemma
  assumes e: "\<exists> x. P \<and> Q(x)"
  shows "P \<and> (\<exists> x. Q(x))"
proof (rule conjI)
  show "P"
    proof -
    from e obtain x where p: "P \<and> Q(x)" by (rule exE)
    from p show "P" by (rule conjunct1)
    qed
  show "\<exists> y. Q(y)"
    proof -
    from e obtain x where p: "P \<and> Q(x)" by (rule exE)
    from p have q: "Q(x)" by (rule conjunct2)
    from q show "\<exists> y. Q(y)" by (rule exI)
    qed
qed

text {* 
  \begin{definicion}[Ejemplo de definición existencial]
  El número natural $x$ divide al número natural $y$ si existe un
  natural $k$ tal que @{text "k \<times> x = y"}. Se representa por @{text "x | y"}.
  \end{definicion}
*}

definition divide :: "nat \<Rightarrow> nat \<Rightarrow> bool" ("_ | _" [80,80] 80) where
  "x | y \<equiv> \<exists> k. k*x = y"

text {* 
  \begin{nota}[Ejemplo de activación automática de regla de simplificación]
    La definición de \verb|divide| se añade a las reglas de
    simplificación. 
  \end{nota}
*}

declare divide_def[simp]

text {* 
  \begin{lema}[Transitividad de la divisibilidad]
  Sean $a$, $b$ y $c$ números naturales. Si $a$ es divisible por $b$ y $b$ es
  divisible por $c$, entonces $a$ es divisible por $c$.
  \end{lema}
*}

lemma divide_trans: 
  fixes a b c :: "nat"
  assumes ab: "a | b" and bc: "b | c"
  shows "a | c"
proof simp
  from ab obtain m where m: "m*a = b" by auto
  from bc obtain n where n: "n*b = c" by auto
  from m n have "m*n*a = c" by auto
  thus "\<exists> k. k*a = c" by (rule exI)
qed

text {* 
  \begin{nota}[Método \emph{auto}]
  En el lema anterior es la primera vez que se usa el método automático
  (\textbf{by} \emph{auto}).   
  \end{nota}

  \begin{lema}[CNS de divisibilidad]
  Sean $a$ y $b$ dos números naturales. Entonces $a$ es divisible por $b$ syss
  el resto de dividir $a$ entre $b$ es cero.  
  \end{lema}
*}

lemma CNS_divisibilidad: 
  "(a | b) = (b mod a = 0)" 
by auto

section {* Razonamiento ecuacional *}

(* Pendiente de añadir de la tesis de Wenzel sección 4.1.3 (p. 71). *)



text {* 
  \begin{nota}[Elementos para el razonamiento ecuacional]
  El razonamiento ecuacional se realiza de manera más concisa usando la
  combinación de \textbf{also} (además) y \textbf{finally} (finalmente).
  \end{nota}
  
  \begin{lema}[Ejemplo de razonamiento ecuacional]
  Si $a=b$, $b=c$ y $c=d$, entonces $a=d$.  
  \end{lema}
*}
lemma 
  assumes 1: "a = b" and 2: "b = c" and 3: "c = d"
  shows "a = d"
proof -
  have "a = b" by (rule 1)
  also have "\<dots> = c" by (rule 2)
  also have "\<dots> = d" by (rule 3)
  finally show "a = d" .
qed

text {* 
  \begin{nota}[Demostración automática con la maza]
    El lema anterior puede demostrarse automáticamente con la maza
    (``sledgehammer''). 
  \end{nota}
*}

lemma 
  assumes 1: "a = b" and 2: "b = c" and 3: "c = d"
  shows "a = d"
proof -
  show "a=d" by (metis 1 2 3)
qed

(*<*)
end
(*>*)
