header {* Distinción de casos e inducción *}

(*<*)
theory Cap_3
imports Main LaTeXsugar OptionalSugar
begin
(*>*)

section {* Razonamiento por distinción de casos *}

subsection {* Distinción de casos booleanos *}

text {*
  \begin{ejemplo}[Demostración por distinción de casos booleanos]
    \mbox{} \newline \hspace*{1cm}
    @{text "\<not>A \<or> A"}
  \end{ejemplo}
*}

lemma "\<not>A \<or> A" 
proof cases
  assume "A" thus ?thesis ..
next
  assume "\<not>A" thus ?thesis ..
qed

text {*
  \begin{ejemplo}[Demostración por distinción de casos booleanos nominados]
    \mbox{} \newline \hspace*{1cm}
    @{text "\<not>A \<or> A"}
  \end{ejemplo}
*}

lemma "\<not>A \<or> A" 
proof (cases "A")
  case True thus ?thesis ..
next
  case False thus ?thesis .. 
qed

text {*
  \begin{nota}[El método \emph{cases} sobre una fórmula]
    \mbox{} 
    \begin{enumerate}
    \item{El método \emph{(cases F)} es una abreviatura de la aplicación de la
      regla \newline \hspace*{1cm} 
      @{term "\<lbrakk>F \<Longrightarrow> Q; \<not> F \<Longrightarrow> Q\<rbrakk> \<Longrightarrow> Q"}}.   
    \item{\textbf{assume} \emph{True} es una abreviatura de $F$.}
    \item{\textbf{assume} \emph{False} es una abreviatura de $\neg F$.}
    \item{Ventajas de \emph{cases} con nombre: reduce la escritura de la fórmula
      y es independiente del orden de los casos.}
    \end{enumerate}
  \end{nota} *}

subsection {* Distinción de casos sobre otros tipos de datos *}

text {*
  \begin{lema}[Distinción de casos sobre listas]
    La longitud del resto de una lista es la longitud de la lista menos 1.
  \end{lema}
*}

lemma "length(tl xs) = length xs - 1" 
proof (cases xs)
  case Nil thus ?thesis by simp
next
  case Cons thus ?thesis by simp 
qed

text {*
  \begin{nota}[Distinción de casos sobre listas]
    \mbox{} 
    \begin{enumerate}
    \item{El método de distinción de casos se activa con (@{text "cases xs"})
        donde @{text "xs"} es del tipo lista.} 
    \item{\textbf{case} \emph{Nil} es una abreviatura de 
        \textbf{assume} \emph{Nil: xs =[]}.}
    \item{\textbf{case} \emph{Cons} es una abreviatura de 
        \textbf {fix} \verb|? ??| \textbf{assume} 
        \verb|Cons: xs = ? # ??|, donde \verb|?| y \verb|??| son
        variables anónimas.} 
    \end{enumerate}
  \end{nota} 

  \begin{lema}[Ejemplo de análisis de casos]
    El resultado de eliminar los $n+1$ primeros elementos de @{term "xs"} es el
    mismo que eliminar los $n$ primeros elementos del resto de  @{term "xs"}. 
  \end{lema}
*}

lemma "drop (n + 1) xs = drop n (tl xs)"
proof (cases xs)
  case Nil thus "drop (n + 1) xs = drop n (tl xs)" by simp
next
  case Cons thus "drop (n + 1) xs = drop n (tl xs)" by simp
qed

text {*
  \begin{nota}[La función \emph{drop}]
  La función \textit{drop} está definida en la teoría \verb|List| de forma que
  @{term "(drop n xs)"} es la lista obtenida eliminando en @{term "xs"} los 
  @{term "n"} primeros elementos. Su definición es la siguiente 
  \begin{equation}
  \begin{array}{lcl}
  @{thm_style lhs drop_Nil } &=& @{thm_style rhs drop_Nil } \\
  @{thm_style lhs drop_Cons } &=& @{thm_style rhs drop_Cons } 
  \end{array}\notag
  \end{equation}
  \end{nota} *}

section {* Inducción matemática *}

text {*
  \begin{nota}[Principio de inducción matemática]
    Para demostrar una propiedad $P$ para todos los números naturales basta
    probar que el $0$ tiene la propiedad $P$ y que si $n$ tiene la propiedad
    $P$, entonces $n+1$ también la tiene.
    \begin{equation}\notag
      @{thm [mode=Rule] nat.induct}
    \end{equation}
  \end{nota}
  
  \begin{nota}[Ejemplo de demostración por inducción]
    Usaremos el principio de inducción matemática para demostrar que
    \begin{equation}\notag
      1 + 3 + \cdots + (2n - 1) = n^2
    \end{equation}
  \end{nota}

  \begin{definicion}[Suma de los primeros impares] 
    @{term "(suma_impares n)"} es la suma de los @{term "n"} primeros números
    impares.    
  \end{definicion}
*}

primrec suma_impares :: "nat \<Rightarrow> nat" where
  "suma_impares 0 = 0"
| "suma_impares (Suc n) = (2 * (Suc n) - 1) + suma_impares n"

text {* 
  \begin{lema}[Ejemplo de suma de impares]
    La suma de los 3 primeros números impares es 9.
  \end{lema}
*}

lemma "suma_impares 3 = 9"
by (simp add:suma_impares_def)

text {*
  \begin{nota}
    La suma de los 3 primero número impares se puede calcular mediante
  \end{nota}
*}

value "suma_impares 3"

text {* 
  que devuelve  el valor @{text "9"}

  \begin{lema}[Ejemplo de demostración por inducción matemática]\label{ind1}
    La suma de los $n$ primeros números impares es $n^2$.
  \end{lema}

  \begin{nota}
    Demostración automática del lema \ref{ind1}.
  \end{nota}
*}

lemma "suma_impares n = n * n"
by (induct n) simp_all

text {*
  \begin{nota}[Los métodos \emph{induct} y \emph{simp\_all}]
    En la demostración \textbf{by} \emph{(induct n) simp\_all} se aplica
    inducción en $n$ y los dos casos se prueban por simplificación. 
  \end{nota}
*}

text {*
  \begin{nota}
    Demostración con patrones del lema \ref{ind1}.
  \end{nota}
*}

lemma "suma_impares n = n * n" (is "?P n")
proof (induct n)
  show "?P 0" by simp
next
  fix n assume "?P n"
  thus "?P(Suc n)" by simp
qed

text {*
  \begin{nota}[Patrones]
    Cualquier fórmula seguida de (\textbf{is} \emph{patrón}) equipara el patrón
    con la fórmula.
  \end{nota}

  \begin{nota}
    Demostración con patrones y razonamiento ecuacional del lema \ref{ind1}.
  \end{nota}
*}

lemma "suma_impares n = n * n" (is "?P n")
proof (induct n)
  show "?P 0" by simp
next
  fix n assume HI: "?P n"
  have "suma_impares (Suc n) = (2 * (Suc n) - 1) + suma_impares n" by simp
  also have "\<dots> = (2 * (Suc n) - 1) + n * n" using HI by simp
  also have "\<dots> = n * n + 2 * n + 1" by simp
  finally show "?P(Suc n)" by simp
qed

text {*
  \begin{nota}
    Demostración por inducción y razonamiento ecuacional del lema \ref{ind1}.
  \end{nota}
*}

lemma "suma_impares n = n * n"
proof (induct n)
  show "suma_impares 0 = 0 * 0" by simp
next
  fix n assume HI: "suma_impares n = n * n"
  have "suma_impares (Suc n) = (2 * (Suc n) - 1) + suma_impares n" by simp
  also have "\<dots> = (2 * (Suc n) - 1) + n * n" using HI by simp
  also have "\<dots> = n * n + 2 * n + 1" by simp
  finally show "suma_impares (Suc n) = (Suc n) * (Suc n)" by simp
qed

text {* 
  \begin{definicion}[Números pares]
    Un número natural $n$ es par si existe un natural $m$ tal que $n=m+m$. 
  \end{definicion}
*}

definition par ::  "nat \<Rightarrow> bool" where
  "par n \<equiv> \<exists>m. n=m+m"

text {* 
  \begin{lema}[Ejemplo de inducción y existenciales]
  Para todo número natural $n$, se verifica que @{text "n*(n+1)"} es par.
  \end{lema}
*}

lemma 
  fixes n :: "nat"
  shows "par (n*(n+1))"
proof (induct n)
  show "par (0 * (0 + 1))" by (simp add:par_def)
next
  fix n assume "par (n*(n+1))"
  hence "\<exists>m. n*(n+1) = m+m" by (simp add:par_def)
  then obtain m where m: "n*(n+1) = m+m" by (rule exE)
  hence "(Suc n)*((Suc n)+1) = (m+n+1)+(m+n+1)" by auto
  hence "\<exists>m. (Suc n)*((Suc n)+1) = m+m" by (rule exI)
  thus "par ((Suc n)*((Suc n)+1))" by (simp add:par_def)
qed

section {* Inducción estructural *}

text {*
  \begin{nota}[Inducción estructural]
    \mbox{} 
    \begin{itemize}
    \item En Isabelle puede hacerse inducción estructural sobre cualquier tipo
      recursivo.
    \item La inducción matemática es la inducción estructural sobre el tipo de
      los naturales.
    \item El esquema de inducción estructural sobre listas es
      \begin{equation}\notag
        @{thm [mode=Rule] list.induct }
      \end{equation}
    \item Para demostrar una propiedad para todas las listas basta demostrar
      que la lista vacía tiene la propiedad y que al añadir un elemento a una
      lista que tiene la propiedad se obtiene una lista que también tiene la
      propiedad. 
    \end{itemize}
  \end{nota}

  \begin{nota}[Concatenación de listas]
    En la teoría \verb|List.thy| está definida la concatenación de listas (que
    se representa por \verb|@|) como sigue
    \begin{verbatim}
    primrec
      append_Nil:  "[]@ys = ys"
      append_Cons: "(x#xs)@ys = x#(xs@ys)"
    \end{verbatim}
  \end{nota}

  \begin{lema}[Ejemplo de inducción sobre listas] \label{concAsoc}
    La concatenación de listas es asociativa.
  \end{lema}

  \begin{nota}
    Demostración automática de \ref{concAsoc}.
  \end{nota}
*}

lemma conc_asociativa_1: "xs @ (ys @ zs) = (xs @ ys) @ zs"
by (induct xs) simp_all

text {*
  \begin{nota}
    Demostración estructurada de \ref{concAsoc}.
  \end{nota}
*}

lemma conc_asociativa: "xs @ (ys @ zs) = (xs @ ys) @ zs"
proof (induct xs)
  show "[] @ (ys @ zs) = ([] @ ys) @ zs"
  proof -
    have "[] @ (ys @ zs) = ys @ zs" by simp
    also have "\<dots> = ([] @ ys) @ zs" by simp
    finally show ?thesis .
  qed
next
  fix x xs
  assume HI: "xs @ (ys @ zs) = (xs @ ys) @ zs"
  show "(x#xs) @ (ys @ zs) = ((x#xs) @ ys) @ zs"
  proof -
    have "(x#xs) @ (ys @ zs) = x#(xs @ (ys @ zs))" by simp
    also have "\<dots> = x#((xs @ ys) @ zs)" using HI by simp
    also have "\<dots> = (x#(xs @ ys)) @ zs" by simp
    also have "\<dots> = ((x#xs) @ ys) @ zs" by simp
    finally show ?thesis .
  qed
qed

text {* 
  \begin{ejercicio}[Árboles binarios]
    Definir un tipo de dato para los árboles binarios.
  \end{ejercicio}
*}

datatype 'a arbol = Hoja "'a" | Nodo "'a" "'a arbol" "'a arbol"

text {* 
  \begin{ejercicio}[Imagen especular]
    Definir la función @{term "espejo"} que aplicada a un árbol devuelve su
    imagen especular. 
  \end{ejercicio}
*}

primrec espejo :: "'a arbol \<Rightarrow> 'a arbol" where
  "espejo (Hoja a) = (Hoja a)"
| "espejo (Nodo f x y) = (Nodo f (espejo y) (espejo x))"

text {* 
  \begin{ejercicio}[La imagen especular es involutiva] \label{espInv}
    Demostrar que la función @{term "espejo"} es involutiva; es decir, para
    cualquier árbol @{term "t"}, 
    \newline \hspace*{1cm}
    @{term "espejo(espejo(t))=t"}.
  \end{ejercicio}

  \begin{nota}
    Demostración automática de \ref{espInv}.
  \end{nota}
*}

lemma espejo_involutiva_1: "espejo(espejo(t)) = t"
by (induct t) auto

text {*
  \begin{nota}
    Demostración estructurada de \ref{espInv}.
  \end{nota}
*}

lemma espejo_involutiva: "espejo(espejo(t)) = t" (is "?P t")
proof (induct t)
  fix x :: 'a show "?P (Hoja x)" by simp 
next
  fix t1 :: "'a arbol" assume h1: "?P t1"
  fix t2 :: "'a arbol" assume h2: "?P t2"
  fix x :: 'a
  show "?P (Nodo x t1 t2)" 
  proof -
    have "espejo(espejo(Nodo x t1 t2)) = espejo(Nodo x (espejo t2) (espejo t1))"
      by simp
    also have "\<dots> = Nodo x (espejo (espejo t1)) (espejo (espejo t2))" by simp
    also have "\<dots> = Nodo x t1 t2" using h1 h2 by simp 
    finally show ?thesis .
 qed
qed

text {* 
  \begin{ejercicio}[Aplanamiento de árboles]
    Definir la función @{term "aplana"} que aplane los árboles recorriéndolos en
    orden infijo. 
  \end{ejercicio}
*}

primrec aplana :: "'a arbol \<Rightarrow> 'a list" where
  "aplana (Hoja a) = [a]"
| "aplana (Nodo x t1 t2) = (aplana t1)@[x]@(aplana t2)"

text {* 
  \begin{ejercicio}[Aplanamiento de la imagen especular] \label{AplEsp}
    @{term "aplana(espejo t) = rev(aplana t)"}.
  \end{ejercicio}

  \begin{nota}
     Demostración automática de \ref{AplEsp}.
  \end{nota}
*}

lemma "aplana(espejo t) = rev(aplana t)"
by (induct t) auto

text {* 
  \begin{nota}
     Demostración estructurada de \ref{AplEsp}.
  \end{nota}
*}

lemma "aplana(espejo t) = rev(aplana t)" (is "?P t")
proof (induct t)
  fix x :: 'a show "?P (Hoja x)" by simp 
next
  fix t1 :: "'a arbol" assume h1: "?P t1"
  fix t2 :: "'a arbol" assume h2: "?P t2"
  fix x :: 'a
  show "?P (Nodo x t1 t2)" 
  proof -
    have "aplana (espejo (Nodo x t1 t2)) = aplana (Nodo x (espejo t2) (espejo t1))" by simp
    also have "\<dots> = (aplana(espejo t2))@[x]@(aplana(espejo t1))" by simp
    also have "\<dots> = (rev(aplana t2))@[x]@(rev(aplana t1))" using h1 h2 by simp
    also have "\<dots> = rev((aplana t1)@[x]@(aplana t2))" by simp
    also have "\<dots> = rev(aplana (Nodo x t1 t2))" by simp
    finally show ?thesis .
 qed
qed

(*<*)
end
(*>*)

