header {* Patrones de demostración *}

(*<*)
theory Cap_4
imports Main LaTeXsugar OptionalSugar
begin
(*>*)

section {* Demostraciones por casos *} 

text {*
  \begin{nota}[Regla de eliminación de la disyunción]
    \begin{equation}\notag
     (\textit{disjE})\;@{thm [mode=Rule] disjE }
    \end{equation}
  \end{nota}

  \begin{lema}[Ejemplo de demostración por casos]
    \mbox{} \newline \hspace*{1cm}
    @{text "P \<or> Q \<Longrightarrow> Q \<or> P"}
  \end{lema}
*}

lemma disj_conmutativa: "P \<or> Q \<Longrightarrow> Q \<or> P"
proof -
  assume "P \<or> Q" 
  thus "Q \<or> P" 
  proof (rule disjE) 
    assume P 
    thus ?thesis  by (rule disjI2)
  next
    assume Q 
    thus ?thesis by (rule disjI1) 
  qed 
qed 

text {*
  \begin{nota}
  El lema anterior puede demostrarse automáticamente como se muestra a
  continuación. 
  \end{nota}
*}

lemma disj_conmutativa_auto: "P \<or> Q \<Longrightarrow> Q \<or> P"
by auto

section {* Negación *}

text {*
  \begin{nota}[Reglas de la negación]
    \begin{equation}\notag
     (\textit{notI})\;@{thm [mode=Rule] notI }\qquad
     (\textit{notE})\;@{thm [mode=Rule] notE }
    \end{equation}
  \end{nota}

  \begin{lema}[Ejemplo de demostración con negaciones]
    Si @{text "x\<^bsup>2\<^esup>+y=13"} e @{text "y \<noteq> 4"}, entonces @{text "x \<noteq> 3"}.
  \end{lema}
*}

lemma 
  fixes x :: "nat"
  assumes 1: "x * x + y = 13" 
      and 2: "y \<noteq> 4"
  shows "x \<noteq> 3"
proof (rule notI)
  assume "x = 3"
  with 1 have "y = 4" by simp
  with 2 show "False" by (rule notE)
qed

lemma 
  fixes x :: "nat"
  assumes 1: "x * x + y = 13" 
      and 2: "y \<noteq> 4"
  shows "x \<noteq> 3"
proof (rule notI)
  assume "x = 3"
  with 1 2 show "False" by auto
qed

text {*
  \begin{nota}
    El lema anterior puede demostrarse automáticamente como se muestra a
    continuación. 
  \end{nota}
*}

lemma 
  fixes x :: "nat"
  assumes 1: "x * x + y = 13" 
      and 2: "y \<noteq> 4"
  shows "x \<noteq> 3"
using assms
by auto


section {* Contradicciones *}

text {*
  \begin{nota}[Regla de contradicción]
    \begin{equation}\notag
     (\textit{FalseE})\;@{thm [mode=Rule] FalseE }
    \end{equation}
  \end{nota}

  \begin{lema}[Ejemplo de uso de la regla de contradicción]
    Si $1=2$, entonces $3=7$.
  \end{lema}
*}

lemma 
  assumes "1 = (2::nat)"
  shows "3 = (7::nat)"
proof - 
  have "False" using assms by simp
  thus "3 = (7::nat)" by (rule FalseE)
qed

text {*
  \begin{nota}
    El lema puede demostrarse automáticamente, como sigue.
  \end{nota}
*}

lemma 
  assumes "1 = (2::nat)"
  shows "3 = (7::nat)"
using assms
by auto

text {* 
  \begin{lema}[Ejemplo de demostración por casos y contradicción]
    \mbox{} \newline \hspace*{1cm}
    @{text "{\<not>P, P \<or> Q} \<turnstile> Q"}.
  \end{lema}
*}

lemma disjCE:
  assumes "\<not>P" and "P \<or> Q"
  shows "Q"
using `P \<or> Q`
proof (rule disjE)
  assume "P"
  thus "Q" using `\<not>P` by contradiction
next
  assume "Q"
  thus "Q" by assumption
qed

section {* Equivalencias *}

text {*
  \begin{nota}[Reglas de equivalencia]
    \begin{equation}\notag
     (\textit{iffI})\;@{thm [mode=Rule] iffI }\qquad
     (\textit{iffD1})\;@{thm [mode=Rule] iffD1 }\qquad
     (\textit{iffD2})\;@{thm [mode=Rule] iffD2 }\qquad
    \end{equation}
  \end{nota}

  \begin{lema}[Ejemplo de introducción de equivalencia]
    La fórmula 
    \newline\hspace*{1cm}
    @{text "(R \<longrightarrow> C) \<and> (S \<longrightarrow> C))"} 
    \newline
    es equivalente a 
    \newline\hspace*{1cm}
    @{text "R \<or> S \<longrightarrow> C"}.
  \end{lema}
*}

lemma "((R \<longrightarrow> C) \<and> (S \<longrightarrow> C)) = (R \<or> S \<longrightarrow> C)"
proof (rule iffI)
  assume "(R \<longrightarrow> C) \<and> (S \<longrightarrow> C)"
  thus "R \<or> S \<longrightarrow> C" by blast
next
  assume "R \<or> S \<longrightarrow> C"
  thus "(R \<longrightarrow> C) \<and> (S \<longrightarrow> C)" by blast
qed

text {* 
  \begin{nota}[El método \emph{blast}]
    En la demostración anterior es la primera vez que se usa el método de
    razonamiento automático \emph{blast}.
  \end{nota}

  \begin{nota}
    El lema anterior puede demostrarse automáticamente como se muestra a
    continuación. 
  \end{nota}
*}

lemma "((R \<longrightarrow> C) \<and> (S \<longrightarrow> C)) = (R \<or> S \<longrightarrow> C)"
by auto

text {*
  \begin{lema}[Ejemplo de eliminación de equivalencia]
    \mbox{}
    \begin{enumerate}
      \item @{text "A \<longleftrightarrow> B, A \<turnstile> B"}
      \item @{text "A \<longleftrightarrow> B, B \<turnstile> A"}
    \end{enumerate}
  \end{lema}
*}

lemma assumes "A = B" and "A" shows "B"
using assms 
by (rule iffD1)

lemma assumes "A = B" and "B" shows "A"
using assms
by (rule iffD2)

(*<*)
end
(*>*)

