header {* Conjuntos, funciones y relaciones *}

(*<*)
theory Cap_7
imports Main LaTeXsugar OptionalSugar 
begin
(*>*)

text {*
  \mbox{} \hfill
  \begin{tabular}{l}
    \emph{No busquéis el significado, buscad el uso.} \\
    {\sc L. Wittgenstein}
  \end{tabular}
*}

section {* Conjuntos *}

subsection {* Operaciones con conjuntos *}

text {*
  \begin{nota}
    La teoría elemental de conjuntos es \texttt{HOL/Set.thy}.
  \end{nota}

  \begin{nota}
    En un conjunto todos los elemento son del mismo tipo (por ejemplo, del tipo
    @{term "\<tau>"}) y el conjunto tiene tipo (en el ejemplo, @{term "\<tau> set"}). 
  \end{nota}

  \begin{nota}[Reglas de la intersección]
    \mbox{}
    \begin{itemize}
      \item @{text "\<lbrakk>c \<in> A; c \<in> B\<rbrakk> \<Longrightarrow> c \<in> A \<inter> B"} \hfill \textit{(IntI)}
      \item @{text "c \<in> A \<inter> B \<Longrightarrow> c \<in> A"} \hfill \textit{(IntD1)}
      \item @{text "c \<in> A \<inter> B \<Longrightarrow> c \<in> B"} \hfill \textit{(IntD2)}
    \end{itemize}
  \end{nota}

  \begin{nota}
    Propiedades del complementario:
    \begin{itemize}
      \item @{text "(c \<in> - A) = (c \<notin> A)"} \hfill (\textit{Compl\_iff})
      \item @{text "- (A \<union> B) = - A \<inter> - B"} \hfill (\textit{Compl\_Un})
    \end{itemize}
  \end{nota}

  \begin{nota}
    El conjunto \textbf{vacío} se representa por @{text "{}"} y el
    \textbf{universal} por @{text "UNIV"}. 
  \end{nota}

  \begin{nota}
    Propiedades de la \textbf{diferencia} y del complementario:
    \begin{itemize}
      \item @{text "A \<inter> (B - A) = {}"} \hfill (\textit{Diff\_disjoint})
      \item @{text "A \<union> - A = UNIV"} \hfill (\textit{Compl\_partition})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Reglas de la relación de \textbf{subconjunto}:
    \begin{itemize}
      \item @{text "(\<And>x. x \<in> A \<Longrightarrow> x \<in> B) \<Longrightarrow> A \<subseteq> B"} \hfill (\textit{subsetI})
      \item @{text "\<lbrakk>A \<subseteq> B; c \<in> A\<rbrakk> \<Longrightarrow> c \<in> B"} \hfill (\textit{subsetD})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Ejemplo trivial.
  \end{nota}
*}

lemma "(A \<union> B \<subseteq> C) = (A \<subseteq> C \<and> B \<subseteq> C)"
by blast

text {* 
  \begin{nota}
    Otro ejemplo trivial.
  \end{nota}
*}

lemma "(A \<subseteq> -B) = (B \<subseteq> -A)"
by blast

text {*
  \begin{nota}
  Principio de extensionalidad de conjuntos:
  \begin{itemize}
  \item @{text "(\<And>x. (x \<in> A) = (x \<in> B)) \<Longrightarrow> A = B"} \hfill (\textit{set\_ext})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Reglas de la \textbf{igualdad} de conjuntos:
  \begin{itemize}
  \item @{text "\<lbrakk>A \<subseteq> B; B \<subseteq> A\<rbrakk> \<Longrightarrow> A = B"} \hfill (\textit{equalityI})
  \item @{text "\<lbrakk>A = B; \<lbrakk>A \<subseteq> B; B \<subseteq> A\<rbrakk> \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"} \hfill (\textit{equalityE})
  \end{itemize}
  \end{nota}

  \begin{lema}[Analogía entre intersección y conjunción]
    @{term "x \<in> A \<inter> B"} syss @{term "x \<in> A"} y @{term "x \<in> B"}.
  \end{lema}
*}

lemma "(x \<in> A \<inter> B) = (x \<in> A \<and> x \<in> B)" 
by simp

text {*
  \begin{lema}[Analogía entre unión y disyunción]
    @{term "x \<in> A \<union> B"} syss @{term "x \<in> A"} ó @{term "x \<in> B"}.
  \end{lema}
*}

lemma "(x \<in> A \<union> B) = (x \<in> A \<or> x \<in> B)" 
by simp

text {*
  \begin{lema}[Analogía entre subconjunto e implicación]
    @{term "(A \<subseteq> B)"} syss para todo @{term "x"}, si @{term "x \<in> A"} entonces
    @{term "x \<in> B"}.
  \end{lema}
*}

lemma "(A \<subseteq> B) = (\<forall> x. x \<in> A \<longrightarrow> x \<in> B)" 
by auto

text {*
  \begin{lema}[Analogía entre complementario y negación]
    $x$ pertenece al complementario de $A$ syss $x$ no pertenece a $A$.
  \end{lema}
*}

lemma "(x \<in> -A) = (x \<notin> A)" 
by simp

subsection {* Notación de conjuntos finitos *}

text {*
  \begin{nota}
  La teoría de conjuntos finitos es \texttt{HOL/Finite\_Set.thy}.
  \end{nota}

  \begin{nota}
  Los conjuntos finitos se definen por inducción a partir de las siguientes
  reglas inductivas:
  \begin{itemize}
  \item{El conjunto vacío es un conjunto finito.
  \newline \hspace*{1cm}
  @{text "finite {}"} \hfill (\textit{emptyI})}
  \item{Si se le añade un elemento a un conjunto finito se obtiene otro
  conjunto finito.
  \newline \hspace*{1cm}
  @{text "finite A \<Longrightarrow> finite (insert a A)"} \hfill (\textit{insertI})}
  \end{itemize}
  \end{nota}

  \begin{nota}
  En la notación matemática, las reglas anteriores se representan como sigue:
  \begin{itemize}
  \item{El conjunto vacío es un conjunto finito.
  \newline \hspace*{1cm}
  @{term "finite {}"} \hfill (\textit{emptyI}) }
  \item{Si se le añade un elemento a un conjunto finito se obtiene otro
  conjunto finito.
  \newline \hspace*{1cm}
  @{term "finite A \<Longrightarrow> finite (insert a A)"} \hfill (\textit{insertI}) }
  \end{itemize}
  \end{nota}

  \begin{ejemplo}
  Ejemplos de conjuntos finitos.
  \end{ejemplo}
  
  \hrule *}

lemma 
  "insert 2 {} = {2} \<and>
  insert 3 {2} = {2,3} \<and>
  insert 2 {2,3} = {2,3} \<and>
  {2,3} = {3,2,3,2,2}"
by auto

text {*
  \begin{nota}
  Los conjuntos finitos se representan con la notación conjuntista habitual:
  los elementos entre llaves y separados por comas.
  \end{nota}

  \begin{nota}
  Ejemplo trivial.
  \end{nota}
*}

lemma "{a,b} \<union> {c,d} = {a,b,c,d}" 
by blast

text {*
  \begin{nota}
  Conjetura falsa.
  \end{nota}
*}

lemma "{a,b} \<inter> {b,c} = {b}" 
refute
oops

text {*
  \begin{nota}
  Conjetura corregida.
  \end{nota}
*}

lemma "{a,b} \<inter> {b,c} = (if a=c then {a,b} else {b})"
by auto

text {*
  \begin{nota}[Sumas y productos de conjuntos finitos]
  Se pueden definir la suma y el producto de los elementos de un conjunto
  finito mediante las siguientes funciones:
  \begin{itemize}  
   \item @{text "setsum"} tal que @{text "(setsum f A)"} es la suma de la
    aplicación de @{text f} a los elementos del conjunto finito @{text A}, 
   \item @{text "setprod"} tal que @{text "(setprod f A)"} es producto de la
    aplicación de @{text f} a los elementos del conjunto finito @{text A}, 
   \item @{text "\<Sum>"} tal que @{text "\<Sum>A"} es la suma de los elementos del
    conjunto finito @{text A},
   \item @{text "\<Prod>"} tal que @{text "\<Prod>A"} es el producto de los elementos del
    conjunto finito @{text A}.
  \end{itemize}
  \end{nota}

  \begin{definicion}[Ejemplos de definiciones recursivas sobre conjuntos
  finitos] Sea $A$ un conjunto finito de números naturales.
  \begin{itemize}
  \item{@{text "sumaConj A"} es la suma de los elementos $A$.}
  \item{@{text "productoConj A"} es el producto de los elementos de $A$.}
  \item{@{text "sumaCuadradosconj A"} es la suma de los cuadrados de los
  elementos $A$.} 
  \end{itemize}
  \end{definicion}
*}

definition sumaConj :: "nat set \<Rightarrow> nat" where
  "sumaConj S \<equiv> \<Sum>S"

definition productoConj :: "nat set \<Rightarrow> nat" where
  "productoConj S \<equiv> \<Prod>S"

definition sumaCuadradosConj :: "nat set \<Rightarrow> nat" where
  "sumaCuadradosConj S \<equiv> setsum (\<lambda> x. x*x) S"

text {*
  \begin{nota}
  Para simplificar lo que sigue, declaramos las anteriores definiciones como
  reglas de simplificación.
  \end{nota}
*}

declare sumaConj_def[simp]
declare productoConj_def[simp]
declare sumaCuadradosConj_def[simp]

text {*
  \begin{ejemplo}
  Ejemplos de evaluación de las anteriores definiciones recursivas.
  \end{ejemplo}
*}

lemma 
  "sumaConj {1,2,3,4} = 10 \<and>
  productoConj {1,2,3} = productoConj {3,2} \<and> 
  sumaCuadradosConj {1,2,3,4} = 30"
by simp

text {*
  \begin{nota}[Inducción sobre conjuntos finitos]
  Para demostrar que todos los conjuntos finitos tienen una propiedad $P$ basta
  probar que
  \begin{enumerate}
  \item{El conjunto vacío tiene la propiedad $P$.}
  \item{Si a un conjunto finito que tiene la propiedad $P$ se le añade un nuevo
  elemento, el conjunto obtenido sigue teniendo la propiedad $P$.}
  \end{enumerate}
  En forma de regla
  \newline \hspace*{1cm}
  % @{text "\<lbrakk>finite F; P {}; \<And>x F. \<lbrakk>finite F; x \<notin> F; P F\<rbrakk> \<Longrightarrow> 
  %         P ({x} \<union> F)\<rbrakk> \<Longrightarrow> P F"}
  (\textit{finite\_induct}) \qquad @{thm [mode=Rule] finite_induct}
  \end{nota}

  \begin{lema}[Ejemplo de inducción sobre conjuntos finitos]
  Sea $S$ un conjunto finito de números naturales. Entonces todos los elementos
  de $S$ son menores o iguales que la suma de los elementos de $S$.
  \end{lema}
  
  \textbf{Demostración automática:}
  \vspace*{1ex}
*}

lemma "finite S \<Longrightarrow> \<forall>x\<in>S. x \<le> sumaConj S"
by (induct rule: finite_induct) auto

text {*
  \textbf{Demostración estructurada:}
  \vspace*{1ex}
*}

lemma sumaConj_acota: "finite S \<Longrightarrow> \<forall>x\<in>S. x \<le> sumaConj S"
proof (induct rule: finite_induct)
  show "\<forall>x\<in>{}. x \<le> sumaConj {}" by simp
next
  fix x and F
  assume fF: "finite F" 
     and xF: "x \<notin> F" 
     and HI: "\<forall> x\<in>F. x \<le> sumaConj F"
  show "\<forall>y\<in>insert x F. y \<le> sumaConj (insert x F)"
  proof 
    fix y 
    assume "y \<in> insert x F"
    show "y \<le> sumaConj (insert x F)"
    proof (cases "y = x")
      assume "y = x"
      hence "y \<le> x + (sumaConj F)" by simp
      also have "\<dots> = sumaConj (insert x F)" using fF xF by simp
      finally show ?thesis .
    next
      assume "y \<noteq> x"
      hence "y \<in> F" using `y \<in> insert x F` by simp
      hence "y \<le> sumaConj F" using HI by blast
      also have "\<dots> \<le> x + (sumaConj F)" by simp
      also have "\<dots> = sumaConj (insert x F)" using fF xF by simp
      finally show ?thesis .
    qed
  qed
qed

subsection {* Definiciones por comprensión *}

text {*
  \begin{nota}
    El conjunto de los elementos que cumple la propiedad @{text P} se
    representa por @{text "{x. P}"}. 
  \end{nota}

  \begin{nota}
    Reglas de comprensión (relación entre colección y pertenencia):
    \begin{itemize}
      \item @{text "(a \<in> {x. P x}) = P a"} \hfill \textit{(mem\_Collect\_eq)}
      \item @{text "{x. x \<in> A} = A"} \hfill \textit{(Collect\_mem\_eq)}
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Dos ejemplos triviales.
  \end{nota}
*}

lemma "{x. P x \<or> x \<in> A} = {x. P x} \<union> A"
by blast

lemma "{x. P x \<longrightarrow> Q x} = -{x. P x} \<union> {x. Q x}"
by blast

text {*
  \begin{nota}
    Ejemplo con la sintaxis general de comprehensión.
  \end{nota}
*}

lemma 
  "{p*q | p q. p\<in>prime \<and> q\<in>prime} = 
  {z. \<exists>p q. z = p*q \<and> p\<in>prime \<and> q\<in>prime}"
by blast

text {*
  \begin{nota}
   En HOL, la notación conjuntista es azúcar sintáctica:
   \begin{itemize}
   \item @{text "x\<in>A"} es equivalente a @{text "A(x)"}.
   \item @{text "{x. P}"} es equivalente a @{text "\<lambda>x. P"}.
   \end{itemize}
  \end{nota}
*}

text {*
  \begin{definicion}[Ejemplo de definición por comprensión]
    El conjunto de los pares es el de los números $n$ para los que existe un
    $m$ tal que $n = 2*m$.
  \end{definicion}
*}

definition Pares :: "nat set" where
  "Pares \<equiv> { n. \<exists> m. n = 2*m }"

text {*
  \begin{ejemplo}
    Los números 2 y 34 son pares.
  \end{ejemplo}
*}

lemma 
  "2 \<in> Pares \<and>
  34 \<in> Pares" 
by (simp add: Pares_def)

text {*
  \begin{definicion}
    El conjunto de los impares es el de los números $n$ para los que existe un
    $m$ tal que $n = 2*m + 1$.
  \end{definicion}
*}

definition Impares :: "nat set" where
  "Impares \<equiv> { n. \<exists> m. n = 2*m + 1 }"

text {*
  \begin{lema}[Ejemplo con las reglas de intersección y comprensión]
    El conjunto de los pares es disjunto con el de los impares.
  \end{lema}
*}

lemma "x \<notin> (Pares \<inter> Impares)"
proof 
  fix x assume S: "x \<in> (Pares \<inter> Impares)"
  hence "x \<in> Pares" by (rule IntD1)
  hence "\<exists> m. x = 2 * m" by (simp only: Pares_def mem_Collect_eq)
  then obtain p where p: "x = 2 * p" .. 

  from S have "x \<in> Impares" by (rule IntD2)
  hence "\<exists> m. x = 2 * m + 1" by (simp only: Impares_def mem_Collect_eq)
  then obtain q where q: "x = 2 * q + 1" .. 

  from p and q show "False" by arith
qed

subsection {* Cuantificadores acotados *}

text {*
  \begin{nota}
  Reglas de \textbf{cuantificador universal acotado} (``bounded''):
  \begin{itemize}
    \item @{text "(\<And>x. x \<in> A \<Longrightarrow> P x) \<Longrightarrow> \<forall>x\<in>A. P x"} \hfill (\textit{ballI})
    \item @{text "\<lbrakk>\<forall>x\<in>A. P x; x \<in> A\<rbrakk> \<Longrightarrow> P x"} \hfill (\textit{bspec})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Reglas de \textbf{cuantificador existencial acotado} (``bounded''):
  \begin{itemize}
    \item @{text "\<lbrakk>P x; x \<in> A\<rbrakk> \<Longrightarrow> \<exists>x\<in>A. P x"} 
      \hfill (\textit{bexI})
    \item @{text "\<lbrakk>\<exists>x\<in>A. P x; \<And>x. \<lbrakk>x \<in> A; P x\<rbrakk> \<Longrightarrow> Q\<rbrakk> \<Longrightarrow> Q"} 
      \hfill (\textit{bexE})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Reglas de la \textbf{unión indexada}:
  \begin{itemize}
    \item @{text "(b \<in> (\<Union>x\<in>A. B x)) = (\<exists>x\<in>A. b \<in> B x)"} 
      \hfill (\textit{UN\_iff})
    \item @{text "\<lbrakk>a \<in> A; b \<in> B a\<rbrakk> \<Longrightarrow> b \<in> (\<Union>x\<in>A. B x)"} 
      \hfill (\textit{UN\_I})
    \item @{text "\<lbrakk>b \<in> (\<Union>x\<in>A. B x); \<And>x. \<lbrakk>x \<in> A; b \<in> B x\<rbrakk> \<Longrightarrow> R\<rbrakk> \<Longrightarrow> R"} 
      \hfill (\textit{UN\_E})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Reglas de la \textbf{unión de una familia}:
  \begin{itemize}
    \item @{text "\<Union>S = (\<Union>x\<in>S. x)"} 
       \hfill (\textit{Union\_def})
    \item @{text "(A \<in> \<Union>C) = (\<exists>X\<in>C. A \<in> X)"} 
       \hfill (\textit{Union\_iff})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Reglas de la \textbf{intersección indexada}:
  \begin{itemize}
    \item @{text "(b \<in> (\<Inter>x\<in>A. B x)) = (\<forall>x\<in>A. b \<in> B x)"} 
      \hfill (\textit{INT\_iff})
    \item @{text "(\<And>x. x \<in> A \<Longrightarrow> b \<in> B x) \<Longrightarrow> b \<in> (\<Inter>x\<in>A. B x)"} 
      \hfill (\textit{INT\_I})
    \item @{text "\<lbrakk>b \<in> (\<Inter>x\<in>A. B x); b \<in> B a \<Longrightarrow> R; a \<notin> A \<Longrightarrow> R\<rbrakk> \<Longrightarrow> R"} 
      \hfill (\textit{INT\_E})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Reglas de la \textbf{intersección de una familia}:
  \begin{itemize}
    \item @{text "\<Inter>S = (\<Inter>x\<in>S. x)"} 
      \hfill (\textit{Inter\_def})
    \item @{text "(A \<in> \<Inter>C) = (\<forall>X\<in>C. A \<in> X)"} 
      \hfill (\textit{Inter\_iff})
  \end{itemize}
  \end{nota}

  \begin{nota}
  Abreviaturas:
  \begin{itemize}
    \item @{text "Collect P"} es lo mismo que @{text "{x. P}"}.
    \item @{text "All P"} es lo mismo que @{text "\<forall>x. P x"}.
    \item @{text "Ex P"} es lo mismo que @{text "\<exists>x. P x"}.
    \item @{text "Ball P"} es lo mismo que @{text "\<forall>x\<in>A. P x"}.
    \item @{text "Bex P"} es lo mismo que @{text "\<exists>x\<in>A. P x"}.
  \end{itemize}
  \end{nota}
  *}

subsection {* Conjuntos finitos y cardinalidad *}

text {*
  \begin{nota}
  El número de elementos de un conjunto finito $A$ es el cardinal de $A$ y se
  representa por @{text "card A"}.
  \end{nota}

  \begin{ejemplo}
  Ejemplos de cardinales de conjuntos finitos.
  \end{ejemplo}
*}

lemma 
  "card {} = 0 \<and>
  card {4} = 1 \<and>
  card {4,1} = 2 \<and>
  x \<noteq> y \<Longrightarrow> card {x,y} = 2" 
by simp

text {* 
  \begin{nota}
  Propiedades de cardinales:
  \begin{itemize}
    \item Cardinal de la unión de conjuntos finitos: \newline
      @{text "\<lbrakk>finite A; finite B\<rbrakk> 
              \<Longrightarrow> card A + card B = card (A \<union> B) + card (A \<inter> B)"} 
      \hfill (\textit{card\_Un\_Int})
    \item Cardinal del conjunto potencia: \newline
      @{text "finite A \<Longrightarrow> card (Pow A) = 2 ^ card A"} 
      \hfill (\textit{card\_Pow})
  \end{itemize}
  \end{nota} *}

section {* Funciones *}

text{* La teoría de funciones es @{text "HOL/Fun.thy"}. *}

subsection {* Nociones básicas de funciones *}

text {*
  \begin{nota}
    Principio de \textbf{extensionalidad para funciones}:
    \begin{itemize}
    \item @{text "(\<And>x. f x = g x) \<Longrightarrow> f = g"} 
      \hfill (\textit{ext})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Actualización de funciones  
    \begin{itemize}
    \item @{text "(f(x := y)) z = (if z = x then y else f z)"} 
      \hfill (\textit{fun\_upd\_apply})
    \item @{text "f(x := y, x := z) = f(x := z)"} 
      \hfill (\textit{fun\_upd\_upd})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Función identidad
    \begin{itemize}
    \item @{text "id \<equiv> \<lambda>x. x"} \hfill (\textit{id\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Composición de funciones:
    \begin{itemize}
    \item @{text "f \<circ> g = (\<lambda>x. f (g x))"} \hfill (\textit{o\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Asociatividad de la composición:
    \begin{itemize}
    \item @{text "f \<circ> (g \<circ> h) = (f \<circ> g) \<circ> h"} \hfill (\textit{o\_assoc})
    \end{itemize}
  \end{nota}
*}

subsection {* Funciones inyectivas, suprayectivas y biyectivas *}

text {*
  \begin{nota}
    Función \textbf{inyectiva} sobre @{text "A"}:
    \begin{itemize}
    \item @{text "inj_on f A \<equiv> \<forall>x\<in>A. \<forall>y\<in>A. f x = f y \<longrightarrow> x = y"} 
      \hfill (\textit{inj\_on\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    @{text "inj f"} es una abreviatura de @{text "inj_on f UNIV"}.
  \end{nota}

  \begin{nota}
    Función \textbf{suprayectiva}:
    \begin{itemize}
    \item @{text "surj f \<equiv> \<forall>y. \<exists>x. y = f x"} \hfill (\textit{surj\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Función \textbf{biyectiva}:
    \begin{itemize}
    \item @{text "bij f \<equiv> inj f \<and> surj f"} \hfill (\textit{bij\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Propiedades de las \textbf{funciones inversas}:
    \begin{itemize}
    \item @{text "inj f \<Longrightarrow> inv f (f x) = x"} 
      \hfill (\textit{inv\_f\_f})
    \item @{text "surj f \<Longrightarrow> f (inv f y) = y"} 
      \hfill (\textit{surj\_f\_inv\_f})
    \item @{text "bij f \<Longrightarrow> inv (inv f) = f"} 
      \hfill (\textit{inv\_inv\_eq})
    \end{itemize}
  \end{nota}

  \begin{nota}
    \textbf{Igualdad de funciones} (por extensionalidad):
    \begin{itemize}
    \item @{text "(f = g) = (\<forall>x. f x = g x)"} 
      \hfill (\textit{expand\_fun\_eq})
    \end{itemize}
  \end{nota}

  \begin{lema}
    Una función inyectiva puede cancelarse en el lado izquierdo de la
    composición de funciones.
  \end{lema}
*}

lemma 
  assumes "inj f"
  shows "(f \<circ> g = f \<circ> h) = (g = h)"
proof 
  assume "f \<circ> g = f \<circ> h" 
  thus "g = h" using `inj f` by (simp add:expand_fun_eq inj_on_def) 
next
  assume "g = h" 
  thus "f \<circ> g = f \<circ> h" by auto
qed

text {*
  Una demostración más detallada es la siguiente
*}

lemma 
  assumes "inj f"
  shows "(f \<circ> g = f \<circ> h) = (g = h)"
proof 
  assume "f \<circ> g = f \<circ> h"
  show "g = h"
  proof
    fix x
    have "(f \<circ> g)(x) = (f \<circ> h)(x)" using `f \<circ> g = f \<circ> h` by simp
    hence "f(g(x)) = f(h(x))" by simp
    thus "g(x) = h(x)" using `inj f` by (simp add:inj_on_def)
  qed
next
  assume "g = h"
  show "f \<circ> g = f \<circ> h"
  proof
    fix x
    have "(f \<circ> g) x = f(g(x))" by simp
    also have "\<dots> = f(h(x))" using `g = h` by simp
    also have "\<dots> = (f \<circ> h) x" by simp
    finally show "(f \<circ> g) x = (f \<circ> h) x" by simp
  qed
qed

text {*
  Una demostración más automática es la siguiente
*}

lemma 
  assumes "inj f"
  shows "(f \<circ> g = f \<circ> h) = (g = h)"
by (metis Un_UNIV_left assms id_o inj_iff inj_on_Un o_assoc)

text {*
  El desarrollo de la demostración automática es la siguiente
*}

lemma 
  assumes "inj f"
  shows "(f \<circ> g = f \<circ> h) = (g = h)"
proof (neg_clausify)
  assume 0: "(f \<circ> g \<noteq> f \<circ> h) \<or> (g \<noteq> h)"
  assume 1: "(g = h) \<or> (f \<circ> g = f \<circ> h)"
  have 2: "\<And>X1. inj_on f X1" 
    by (metis assms inj_on_Un Un_UNIV_left)
  have 3: "(inv f \<circ> (f \<circ> g) = h) \<or> h = g \<or> \<not> inj f"
    by (metis 1 o_assoc inj_iff id_o)
  have 4: "(inv f \<circ> (f \<circ> g) = h) \<or> h = g" by (metis 2 3)
  have 5: "h = g \<or> \<not> inj f" by (metis id_o o_assoc inj_iff 4)
  have 6: "h = g" by (metis 5 2)
  have 7: "h \<noteq> g" by (metis 6 0)
  show "False" by (metis 6 7)
qed

subsubsection {* Función imagen *}

text {*
  \begin{nota}
    \textbf{Imagen de un conjunto} mediante una función:
    \begin{itemize}
      \item @{text "f ` A = {y. (\<exists>x\<in>A. y = f x)"} 
        \hfill (\textit{image\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Propiedades de la imagen:
    \begin{itemize}
      \item @{text "(f \<circ> g)`r = f`g`r"} 
        \hfill (\textit{image\_compose})
      \item @{text "f`(A \<union> B) = f`A \<union> f`B"} 
        \hfill (\textit{image\_Un})
      \item @{text "inj f \<Longrightarrow> f`(A \<inter> B) = f`A \<inter> f`B"} 
        \hfill (\textit{image\_Int})
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Ejemplos de demostraciones triviales de propiedades de la imagen.
  \end{nota}
*}

lemma "f`A \<union> g`A = (\<Union>x\<in>A. {f x, g x})"
by auto

lemma "f`{(x,y). P x y} = {f(x,y) | x y. P x y}"
by auto

text {*
  \begin{nota}
    El \textbf{rango} de una función (@{text "range f"}) es la imagen del
    universo (@{text "f`UNIV"}). 
  \end{nota}

  \begin{nota}
    \textbf{Imagen inversa de un conjunto}:
    \begin{itemize}
    \item @{text "f -` B \<equiv> {x. f x : B}"} \hfill (\textit{vimage\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Propiedad de la imagen inversa de un conjunto:
    \begin{itemize}
    \item @{text "f -` (-A) = -(f -` A)"} \hfill (\textit{vimage\_Compl})
    \end{itemize}
  \end{nota}
*}

section {* Relaciones *}

subsection {* Relaciones básicas *}

text {*
  \begin{nota}
    La teoría de relaciones es @{text "HOL/Relation.thy"}.
  \end{nota}

  \begin{nota}
    Las relaciones son conjuntos de pares.
  \end{nota}

  \begin{nota}
    \textbf{Relación identidad}:
    \begin{itemize}
      \item @{text "Id \<equiv> {p. EX x. p = (x,x)}"} \hfill (\textit{Id\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    \textbf{Composición de relaciones}:
    \begin{itemize}
      \item @{text "r O s \<equiv> {(x,z). EX y. (x, y) \<in> r & (y, z) \<in> s}"} 
        \hfill (\textit{rel\_comp\_def})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Propiedades:
    \begin{itemize}
      \item @{text "R O Id = R"} \hfill (\textit{R\_O\_Id})
      \item @{text "\<lbrakk>r' \<subseteq> r; s' \<subseteq> s\<rbrakk> \<Longrightarrow> (r' O s') \<subseteq> (r O s)"} 
        \hfill (\textit{rel\_comp\_mono})
    \end{itemize}
  \end{nota}

  \begin{nota}
    \textbf{Imagen inversa de una relación}:
    \begin{itemize}
      \item @{text "((a,b) \<in> r\<^bsup>\<^sup>-1\<^esup>) = ((b,a) \<in> r)"} 
        \hfill (\textit{converse\_iff})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Propiedad de la imagen inversa de una relación:
    \begin{itemize}
      \item @{text "(r O s)\<^bsup>-1\<^esup> = s\<^bsup>-1\<^esup> O r\<^bsup>-1\<^esup>"} 
        \hfill (\textit{converse\_rel\_comp})
    \end{itemize}
  \end{nota}

  \begin{nota}
    \textbf{Imagen} de un conjunto mediante una relación:
    \begin{itemize}
      \item @{text "(b \<in> r``A) = (\<exists>x:A. (x, b) \<in> r)"} 
        \hfill (\textit{Image\_iff})
    \end{itemize}
  \end{nota}

  \begin{nota}
    \textbf{Dominio} de una relación:
    \begin{itemize}
      \item @{text "(a \<in> Domain r) = (\<exists>y. (a, y) \<in> r)"} 
        \hfill (\textit{Domain\_iff})
    \end{itemize}
  \end{nota}

  \begin{nota}
    \textbf{Rango} de una relación:
    \begin{itemize}
      \item @{text "(a \<in> Range r) = (\<exists>y. (y, a) \<in> r)"} 
        \hfill (\textit{Range\_iff})
    \end{itemize}
  \end{nota}
*}

subsection {* Clausura reflexiva y transitiva *}

text {*
  \begin{nota}
    La teoría de la clausura reflexiva y transitiva de una relación es \newline
    @{text "HOL/Transitive_Closure.thy"}.
  \end{nota}

  \begin{nota}
    Potencias de relaciones:
    \begin{itemize}
      \item @{text "R ^^ 0 = Id"}
      \item @{text "R ^^ (Suc n) = (R ^^ n) O R"}
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    La \textbf{clausura reflexiva y transitiva} de la relación @{term "r"} es 
    la menor solución de la ecuación:
    \begin{itemize}
      \item @{text "r\<^sup>* = Id \<union> (r\<^sup>* O r)"} \hfill (\textit{rtrancl\_unfold})
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Propiedades básicas de la clausura reflexiva y transitiva:
    \begin{itemize}
      \item @{text "(a, a) \<in> r\<^sup>*"} 
        \hfill (\textit{rtrancl\_refl})
      \item @{text "p \<in> r \<Longrightarrow> p \<in> r\<^sup>*"} 
        \hfill (\textit{r\_into\_rtrancl})
      \item @{text "\<lbrakk>(a, b) \<in> r\<^sup>*; (b, c) \<in> r\<^sup>*\<rbrakk> \<Longrightarrow> (a, c) \<in> r\<^sup>*"} 
        \hfill (\textit{rtrancl\_trans})
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Inducción sobre la clausura reflexiva y transitiva
    \begin{itemize}
      \item % @{text "\<lbrakk>(a, b) \<in> r\<^sup>*; 
            %          P b; 
            %          \<And>y z. \<lbrakk>(y, z) \<in> r; (z, b) \<in> r\<^sup>*; P z\<rbrakk> \<Longrightarrow> P y\<rbrakk>
            %         \<Longrightarrow> P a"}
            @{thm [mode=Rule] rtrancl_induct} 
            \hfill (\textit{rtrancl\_induct})
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Idempotencia de la clausura reflexiva y transitiva:
    \begin{itemize}
      \item @{text "(r\<^sup>* )\<^sup>* = r\<^sup>*"} 
        \hfill (\textit{rtrancl\_idemp})
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Reglas de introducción de la \textbf{clausura transitiva}:
    \begin{itemize}
      \item @{text "p \<in> r \<Longrightarrow> p \<in> r\<^sup>+"} 
        \hfill (\textit{r\_into\_trancl'})
      \item @{text "\<lbrakk>(a, b) \<in> r\<^sup>+; (b, c) \<in> r\<^sup>+\<rbrakk> \<Longrightarrow> (a, c) \<in> r\<^sup>+"} 
        \hfill (\textit{trancl\_trans})
    \end{itemize}
  \end{nota}
  
  \begin{nota}
    Ejemplo de propiedad:
    \begin{itemize}
      \item @{text "(r\<inverse>)\<^sup>+ = (r\<^sup>+)\<inverse>"} 
        \hfill (\textit{trancl\_converse})
    \end{itemize}
  \end{nota}
*}

subsection {* Una demostración elemental *}

text {*
  \begin{nota}
    El teorema que se desea demostrar es que la clausura reflexiva y
    transitiva conmuta con la inversa (@{text rtrancl_converse}). Para
    demostrarlo introducimos dos lemas auxiliares: @{text rtrancl_converseD} y
    @{text rtrancl_converseI}.
  \end{nota}
*}

lemma rtrancl_converseD: "(x,y) \<in> (r\<inverse>)\<^sup>* \<Longrightarrow> (y,x) \<in> r\<^sup>*"
proof (induct rule:rtrancl_induct)
  show "(x,x) \<in> r\<^sup>*" by (rule rtrancl_refl) 
next
  fix y z
  assume "(x,y) \<in> (r\<inverse>)\<^sup>*" and "(y,z) \<in> r\<inverse>" and "(y,x) \<in> r\<^sup>*"
  show "(z,x) \<in> r\<^sup>*"
  proof (rule rtrancl_trans)
    show "(z,y) \<in> r\<^sup>*" using `(y,z) \<in> r\<inverse>` by simp
  next
    show "(y,x) \<in> r\<^sup>*" using `(y,x) \<in> r\<^sup>*` by simp
  qed
qed

lemma rtrancl_converseI: "(y,x) \<in> r\<^sup>* \<Longrightarrow> (x,y) \<in> (r\<inverse>)\<^sup>*"
proof (induct rule:rtrancl_induct)
  show "(y,y) \<in> (r\<inverse>)\<^sup>*" by (rule rtrancl_refl) 
next
  fix u z
  assume "(y,u) \<in> r\<^sup>*" and "(u,z) \<in> r" and "(u,y) \<in> (r\<inverse>)\<^sup>*"
  show "(z,y) \<in> (r\<inverse>)\<^sup>*"
  proof (rule rtrancl_trans)
    show "(z,u) \<in> (r\<inverse>)\<^sup>*" using `(u,z) \<in> r` by auto
  next
    show "(u,y) \<in> (r\<inverse>)\<^sup>*" using `(u,y) \<in> (r\<inverse>)\<^sup>*` by simp
  qed
qed

theorem rtrancl_converse: "(r\<inverse>)\<^sup>* = (r\<^sup>*)\<inverse>"
proof
  show "(r\<inverse>)\<^sup>* \<subseteq> (r\<^sup>*)\<inverse>" by (auto simp add:rtrancl_converseD)
next
  show "(r\<^sup>*)\<inverse> \<subseteq> (r\<inverse>)\<^sup>*" by (auto simp add:rtrancl_converseI)
qed

text {*
  \begin{nota}
    Puede demostrarse de manera más corta como sigue:
  \end{nota}
*}

theorem "(r\<inverse>)\<^sup>* = (r\<^sup>*)\<inverse>"
by (auto intro: rtrancl_converseI dest: rtrancl_converseD)

section {* Relaciones bien fundamentadas e inducción *}

text {*
  \begin{nota}
    La teoría de las relaciones bien fundamentadas es 
    @{text "HOL/Wellfounded_Relations.thy"}.
  \end{nota}

  \begin{nota}
    La relación--objeto @{text "less_than"} es el orden de los naturales que es
    bien fundamentada:
    \begin{itemize}
      \item @{text "((x,y) \<in> less_than) = (x < y)"} 
        \hfill (\textit{less\_than\_iff})
      \item @{text "wf less_than"} 
        \hfill (\textit{wf\_less\_than})
    \end{itemize}
  \end{nota}

  \begin{nota}
    Notas sobre \textbf{medidas}:
    \begin{itemize}
      \item \textbf{Imagen inversa} de una relación mediante una función:\\
        @{text "inv_image r f \<equiv> {(x,y). (f x,f y) \<in> r"} 
        \hfill @{text "(inv_image_def)"}
      \item Conservación de la buena fundamentación:\\
        @{text "wf r \<Longrightarrow> wf (inv_image r f)"} 
        \hfill @{text "(wf_inv_image)"}
      \item Definición de la \textbf{medida}:\\
        @{text "measure \<equiv> inv_image less_than"} 
        \hfill @{text "(measure_def)"}
      \item Buena fundamentación de la medida:\\
        @{text "wf (measure f)"} 
        \hfill @{text "(wf_measure)"}
    \end{itemize}
  \end{nota}
*}

text {*
  \begin{nota}
    Notas sobre el \textbf{producto lexicográfico}:
    \begin{itemize}
      \item Definición del producto lexicográfico @{text "(lex_prod_def)"}:\\
        @{text "ra <*lex*> rb \<equiv> 
                {((a,b),(a',b')). (a,a') \<in> ra \<or> 
                                  (a = a' \<and> (b,b') \<in> rb)}"}
      \item Conservación de la buena fundamentación:\\ 
        @{text "\<lbrakk>wf ra; wf rb\<rbrakk> \<Longrightarrow> wf (ra <*lex*> rb)"} 
        \hfill @{text "(wf_lex_prod)"}
    \end{itemize}
  \end{nota}

  \begin{nota}
    El orden de multiconjuntos está en la teoría 
    @{text "HOL/Library/Multiset.thy"}.
  \end{nota}

  \begin{nota}
    \textbf{Inducción sobre relaciones bien fundamentadas}:
    \begin{itemize}
      % @{text "\<lbrakk>wf r; \<And>x. (\<And>y. (y,x) \<in> r \<Longrightarrow> P y) \<Longrightarrow> P x\<rbrakk> \<Longrightarrow> P a"}
      \item @{thm [mode=Rule] wf_induct} \hfill @{text "(wf_induct)"}
    \end{itemize}
  \end{nota}
*}

(*<*)
end
(*>*)

